from shutil import copyfile
from pathlib import Path
from tempfile import TemporaryDirectory

from requests.exceptions import ConnectionError
import pandas as pd

from boo.main import download, read_dataframe, inspect
from boo.path import RawFile


DATA_FOLDER = Path(__file__).parent / "data"


def raw_content():
    return (DATA_FOLDER / "raw0.csv").read_text(encoding="cp1251")


def copy(filename, destination_folder):
    src = DATA_FOLDER / filename
    dst = Path(destination_folder) / filename
    copyfile(src, dst)


def test_download():
    try:
        with TemporaryDirectory() as temp_dir:
            download(year=0, directory=temp_dir)
            r = RawFile(year=0, directory=temp_dir)
            assert r.content() == raw_content()
    # test passes if no internet connection available
    except ConnectionError:
        pass



def test_read_dataframe():
    with TemporaryDirectory() as temp_dir:
        copy("raw0.csv", temp_dir)
        df = read_dataframe(year=0, directory=temp_dir)
        assert isinstance(df, pd.DataFrame)
        assert sum(df.cf) == -7032726


def test_inspect_on_existing_data_is_callable(ARGS_YEAR_0):
    year, folder = ARGS_YEAR_0
    inspect(year, folder)


def test_inspect_on_non_existing_data_is_callable():
    with TemporaryDirectory() as temp_dir:
        inspect(2012, directory=temp_dir)



def n_largest(df, col: str, n: int):
    return sort(df, col).head(n)


def contains(df, text):
    ix = df.title.str.lower().str.contains(text.lower()).fillna(False)
    return df[ix]


def sort(df, col):
    return df.sort_values(col, ascending=False)


def region(df, x: int):
    return df[df.region == x]


def inn(df, x: str):
    if isinstance(x, (str, int)):
        ix = str(x)
    else:
        xs = list(map(str, x))
        ix = df.index.isin(xs)
    return df.loc[ix, :]


def industry(df, ok1, ok2=None):
    if ok2 is None:
        return df[df.ok1 == ok1]
    else:
        return df[(df.ok1 == ok1) & (df.ok2 == ok2)]


def dim(df):
    return f"Dataset has {df.shape[0]} rows and {df.shape[1]} columns."



"""File download utility."""
from urllib.request import urlopen

import requests
from tqdm import tqdm

from boo.year import make_url
from boo.helper import as_mb


def curl(path: str, url: str, max_chunk=None):
    r = requests.get(url, stream=True)
    iterable = tqdm(r.iter_content(chunk_size=1024), unit=" k")
    with open(path, "wb") as f:
        i = 0
        for chunk in iterable:
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
            i += 1
            if max_chunk and i >= max_chunk:
                break


def url_content_length(url: str) -> int:
    with urlopen(url) as f:
        return int(f.info()["Content-Length"])


def file_length(year: int):
    return url_content_length(url=make_url(year))


def file_length_mb(year: int):
    return as_mb(file_length(year))



import pytest
from boo.year import make_url


def test_make_url_on_0():
    assert make_url(0)


def test_make_url_on_good_year():
    assert make_url(2012)


def test_make_url_on_bad_year():
    with pytest.raises(ValueError):
        make_url(1990)



import pytest
import os
from pathlib import Path

from boo.path import RawFile,  get_folder


def test_get_folder(ARGS_YEAR_0):
    _, dir0 = ARGS_YEAR_0
    f = get_folder(directory=dir0)
    assert isinstance(f, Path)
    assert str(f) == dir0


def test_get_folder_on_none():
    f = get_folder(directory=None)
    assert f.is_dir()
    path, file = os.path.split(f)
    assert file == ".boo"


def test_size_raw(ARGS_YEAR_0, SIZE_RAW):
    year, folder = ARGS_YEAR_0
    r = RawFile(year, folder)
    assert r.size() == SIZE_RAW > 0  # 11490


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



from boo.errors import WrongYearError

# Note: must manually hardcode new timestamps when new version of data arrives.

TIMESTAMPS = {
    2012: 20190329,
    2013: 20190411,
    2014: 20190411,
    2015: 20190411,
    2016: 20190411,
    2017: 20190423,
    2018: 20191029,
}


def available_years():
    """List available years with datasets."""
    return list(TIMESTAMPS.keys())


def get_timestamp(year: int, timestamps=TIMESTAMPS):
    try:
        return timestamps[year]
    except KeyError:
        raise WrongYearError(year, allowed=available_years())


def rosstat_url(year: int, timestamps=TIMESTAMPS):
    """
    Construct URL similar to
    http://www.gks.ru/opendata/storage/7708234640-bdboo2012/data-20181029t000000-structure-20121231t000000.csv
    using timestamps.
    """
    timestamp = get_timestamp(year)
    return (
        "http://www.gks.ru/opendata/storage/"
        f"7708234640-bdboo{year}/"
        f"data-{timestamp}t000000-structure-"
        f"{year}1231t000000.csv"
    )


def testing_url(year: int):
    return (
        "https://raw.githubusercontent.com/"
        "ru-corporate/boo/master/assets/"
        f"sample{year}.txt"
    )


def make_url(year: int) -> str:
    if year in [0, 1]:  # allowed year numbers for testing
        return testing_url(year)
    return rosstat_url(year)



import pytest
from pathlib import Path
from tempfile import NamedTemporaryFile

from boo.curl import curl, file_length, file_length_mb
from boo.year import make_url


@pytest.fixture
def tempname():  # FIXME: maybe there is a shorter way to make a tempname
    with NamedTemporaryFile() as f:
        name = f.name
    yield name
    if Path(name).exists():
        Path(name).unlink()


def test_curl(tempname):
    n = 3
    curl(path=tempname, url=make_url(2012), max_chunk=n)
    assert Path(tempname).stat().st_size == n * 1024


def test_file_length():
    assert file_length(0) == 11490


def test_file_length_mb():
    assert file_length_mb(0) == 0



from boo.dataframe.canonic import is_numeric_column

# columns types


def numeric_columns(df):
    return list(filter(is_numeric_column, df.columns))


def text_columns(df):
    s = set(df.columns) - set(numeric_columns(df))
    return list(s)


def split_columns(df):
    return text_columns(df), numeric_columns(df), df.columns


# change rows


def is_alive(df):
    return (df.sales != 0) | (df.cf != 0) | (df.profit_before_tax != 0)
    # MAYBE: add large deviations from accounting identity


# change values


def change_numeraire(df, unit):
    """Change unit of account (numeraire), eg thousands to billion.
       Assumes *df* units are thousand rubles.
    """
    text_cols, num_cols, all_cols = split_columns(df)
    return df.loc[:, num_cols].divide(unit).round(1).join(df[text_cols])[all_cols]


def to_bln(df):
    return change_numeraire(df, unit=1_000_000)


def to_mln(df):  # from thousands, default
    return change_numeraire(df, unit=1_000)


# export


def large_companies(df):
    _df = (
        df.loc[is_alive(df), :]
        .query("ta > 1_000_000")
        .sort_values("ta", ascending=False)
    )
    return to_bln(_df)


def medium_companies(df):
    _df = df.query("sales >= 1_000").sort_values("sales", ascending=False)
    return to_mln(_df)


def small_companies(df):
    _df = df.query("sales <= 1_000").sort_values("sales", ascending=False)
    return to_mln(_df)


def shorthand(df):
    return df.rename(
        columns={
            "profit_before_tax": "p",
            "profit_before_tax_lag": "p_lag",
            "tp_capital": "cap",
            "tp_capital_lag": "cap_lag",
        }
    )


def no_lags(df):
    return df[[c for c in df.columns if ("_lag" not in c)]]


class Columns:
    MINIMAL = ["region", "ok1", "ok2", "title"] + [
        "ta",
        "of",
        "sales",
        "profit_before_tax",
        "cf",
    ]
    CF = ["cf_oper", "cf_inv", "cf_fin"]


def minimal_columns(df):
    return shorthand(df[Columns.MINIMAL])


from setuptools import setup, find_packages
from pathlib import Path

from changelog import version_str

# Use script upload.bat to update pipy package.

with open("README.md", encoding="utf-8") as file:
    readme_str = "\n".join(file.readlines())

setup(
    name="boo",
    version=version_str,
    description="Russian corporate reports 2012-2018",
    url="http://github.com/ru-corporate/boo",
    author="Evgeniy Pogrebnyak",
    author_email="e.pogrebnyak@gmail.com",
    license="MIT",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    long_description=readme_str,
    long_description_content_type="text/markdown",
    zip_safe=False,
    install_requires=["requests", "pandas", "tqdm"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Financial and Insurance Industry",
        "Topic :: Office/Business :: Financial",
        "Topic :: Utilities",
    ],
)



"""CSV file load, read and write operations."""

import csv
import os
from tqdm import tqdm

import pandas as pd


def panic(path):
    if os.path.exists(path):
        raise FileExistsError("File already exists: %s" % path)


def yield_rows(path, enc="windows-1251", sep=";"):
    """Emit CSV rows by filename."""
    with open(path, "r", encoding=enc) as csvfile:
        spamreader = csv.reader(csvfile, delimiter=sep)
        for row in spamreader:
            # filter only non-empty rows, prevent error in `cutf(2016)`
            if row:
                yield row


def save_rows(
    path,
    stream,
    column_names=None,
    fmt=dict(lineterminator="\n", quoting=csv.QUOTE_MINIMAL),
    enc="utf-8",
):
    panic(path)
    iterable = tqdm(stream, unit=" lines")
    with open(path, "w", encoding=enc) as file:
        writer = csv.writer(file, **fmt)
        if column_names:
            writer.writerow(column_names)
        writer.writerows(iterable)


def read_df(path, dtypes):
    with open(path, "r", encoding="utf-8") as f:
        return pd.read_csv(f, dtype=dtypes)



from .main import download, read_dataframe, inspect
from .year import make_url, available_years
from .whatis import whatis
from .curl import file_length, file_length_mb
from .okved import all_codes_v2, name_v2
from .dataframe.filter import large_companies, no_lags, shorthand, minimal_columns
